@@ -17,6 +17,8 @@
 #  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 #  USA
 """Classes for working with locally available Debian packages."""
+from __future__ import print_function
+
 import apt
 import apt_inst
 import apt_pkg
@@ -24,43 +26,62 @@ import gzip
 import os
 import sys
 
+try:
+    from typing import Dict, Iterable, List, Optional, Set, Tuple, Union, cast
+    Dict  # pyflakes
+    Iterable  # pyflakes
+    List  # pyflakes
+    Optional  # pyflakes
+    Set  # pyflakes
+    Tuple  # pyflakes
+    Union  # pyflakes
+except ImportError:
+    def cast(typ, obj):  # type: ignore
+        return obj
+    pass
+
 from apt_pkg import gettext as _
-from StringIO import StringIO
+from io import BytesIO
+
 
 class NoDebArchiveException(IOError):
     """Exception which is raised if a file is no Debian archive."""
 
+
 class DebPackage(object):
     """A Debian Package (.deb file)."""
 
     # Constants for comparing the local package file with the version
     # in the cache
-    (VERSION_NONE, 
-     VERSION_OUTDATED, 
-     VERSION_SAME, 
+    (VERSION_NONE,
+     VERSION_OUTDATED,
+     VERSION_SAME,
      VERSION_NEWER) = range(4)
 
     debug = 0
 
     def __init__(self, filename=None, cache=None):
+        # type: (Optional[str], Optional[apt.Cache]) -> None
         if cache is None:
             cache = apt.Cache()
         self._cache = cache
-        self._debfile = None
+        self._debfile = cast(apt_inst.DebFile, None)
         self.pkgname = ""
-        self._sections = {}
-        self._need_pkgs = []
+        self.filename = None  # type: Optional[str]
+        self._sections = {}  # type: Union[Dict[str, str], apt_pkg.TagSection[str]]  # noqa
+        self._need_pkgs = []  # type: List[str]
         self._check_was_run = False
         self._failure_string = ""
-        self._multiarch = None
+        self._multiarch = None  # type: Optional[str]
         if filename:
             self.open(filename)
 
     def open(self, filename):
+        # type: (str) -> None
         """ open given debfile """
         self._dbg(3, "open '%s'" % filename)
         self._need_pkgs = []
-        self._installed_conflicts = set()
+        self._installed_conflicts = set()  # type: Set[str]
         self._failure_string = ""
         self.filename = filename
         self._debfile = apt_inst.DebFile(self.filename)
@@ -68,12 +89,18 @@ class DebPackage(object):
         self._sections = apt_pkg.TagSection(control)
         self.pkgname = self._sections["Package"]
         self._check_was_run = False
-        
+
     def __getitem__(self, key):
+        # type: (str) -> str
         return self._sections[key]
 
+    def __contains__(self, key):
+        # type: (str) -> bool
+        return key in self._sections
+
     @property
     def filelist(self):
+        # type: () -> List[str]
         """return the list of files in the deb."""
         files = []
         try:
@@ -85,34 +112,43 @@ class DebPackage(object):
 
     @property
     def control_filelist(self):
-        """ return the list of files in control.tar.gt """
+        # type: () -> List[str]
+        """ return the list of files in control.tar.gz """
         control = []
         try:
-            self._debfile.control.go(lambda item, data: control.append(item.name))
+            self._debfile.control.go(
+                lambda item, data: control.append(item.name))
         except SystemError:
             return [_("List of control files for '%s' could not be read") %
                     self.filename]
         return sorted(control)
 
-
     # helper that will return a pkgname with a multiarch suffix if needed
-    def _maybe_append_multiarch_suffix(self, pkgname, 
+    def _maybe_append_multiarch_suffix(self, pkgname,
                                        in_conflict_checking=False):
+        # type: (str, bool) -> str
         # trivial cases
+        if ":" in pkgname:
+            return pkgname
         if not self._multiarch:
             return pkgname
         elif self._cache.is_virtual_package(pkgname):
             return pkgname
-        elif (pkgname in self._cache and 
-              self._cache[pkgname].candidate.architecture == "all"):
+        elif (pkgname in self._cache and
+              self._cache[pkgname].candidate is not None and
+              cast(apt.package.Version,
+                   self._cache[pkgname].candidate).architecture == "all"):
             return pkgname
         # now do the real multiarch checking
         multiarch_pkgname = "%s:%s" % (pkgname, self._multiarch)
         # the upper layers will handle this
-        if not multiarch_pkgname in self._cache:
+        if multiarch_pkgname not in self._cache:
+            return multiarch_pkgname
+        multiarch_pkg = self._cache[multiarch_pkgname]
+        if multiarch_pkg.candidate is None:
             return multiarch_pkgname
         # now check the multiarch state
-        cand = self._cache[multiarch_pkgname].candidate._cand
+        cand = multiarch_pkg.candidate._cand
         #print pkgname, multiarch_pkgname, cand.multi_arch
         # the default is to add the suffix, unless its a pkg that can satify
         # foreign dependencies
@@ -120,12 +156,13 @@ class DebPackage(object):
             return pkgname
         # for conflicts we need a special case here, any not multiarch enabled
         # package has a implicit conflict
-        if (in_conflict_checking and 
-            not (cand.multi_arch & cand.MULTI_ARCH_SAME)):
+        if (in_conflict_checking and
+                not (cand.multi_arch & cand.MULTI_ARCH_SAME)):
             return pkgname
         return multiarch_pkgname
 
     def _is_or_group_satisfied(self, or_group):
+        # type: (List[Tuple[str, str, str]]) -> bool
         """Return True if at least one dependency of the or-group is satisfied.
 
         This method gets an 'or_group' and analyzes if at least one dependency
@@ -142,9 +179,11 @@ class DebPackage(object):
             depname = self._maybe_append_multiarch_suffix(depname)
 
             # check for virtual pkgs
-            if not depname in self._cache:
+            if depname not in self._cache:
                 if self._cache.is_virtual_package(depname):
-                    self._dbg(3, "_is_or_group_satisfied(): %s is virtual dep" % depname)
+                    self._dbg(
+                        3, "_is_or_group_satisfied(): %s is virtual dep" %
+                        depname)
                     for pkg in self._cache.get_providing_packages(depname):
                         if pkg.is_installed:
                             return True
@@ -157,17 +196,20 @@ class DebPackage(object):
             # if no real dependency is installed, check if there is
             # a package installed that provides this dependency
             # (e.g. scrollkeeper dependecies are provided by rarian-compat)
-            # but only do that if there is no version required in the 
+            # but only do that if there is no version required in the
             # dependency (we do not supprot versionized dependencies)
             if not oper:
                 for ppkg in self._cache.get_providing_packages(
                     depname, include_nonvirtual=True):
                     if ppkg.is_installed:
-                        self._dbg(3, "found installed '%s' that provides '%s'" % (ppkg.name, depname))
+                        self._dbg(
+                            3, "found installed '%s' that provides '%s'" % (
+                                ppkg.name, depname))
                         return True
         return False
 
     def _satisfy_or_group(self, or_group):
+        # type: (List[Tuple[str, str, str]]) -> bool
         """Try to satisfy the or_group."""
         for dep in or_group:
             depname, ver, oper = dep
@@ -176,7 +218,7 @@ class DebPackage(object):
             depname = self._maybe_append_multiarch_suffix(depname)
 
             # if we don't have it in the cache, it may be virtual
-            if not depname in self._cache:
+            if depname not in self._cache:
                 if not self._cache.is_virtual_package(depname):
                     continue
                 providers = self._cache.get_providing_packages(depname)
@@ -206,20 +248,26 @@ class DebPackage(object):
             or_str += dep[0]
             if ver and oper:
                 or_str += " (%s %s)" % (dep[2], dep[1])
-            if dep != or_group[len(or_group)-1]:
+            if dep != or_group[len(or_group) - 1]:
                 or_str += "|"
-        self._failure_string += _("Dependency is not satisfiable: %s\n") % or_str
+        self._failure_string += _(
+            "Dependency is not satisfiable: %s\n") % or_str
         return False
 
     def _check_single_pkg_conflict(self, pkgname, ver, oper):
+        # type: (str, str, str) -> bool
         """Return True if a pkg conflicts with a real installed/marked pkg."""
         # FIXME: deal with conflicts against its own provides
         #        (e.g. Provides: ftp-server, Conflicts: ftp-server)
-        self._dbg(3, "_check_single_pkg_conflict() pkg='%s' ver='%s' oper='%s'" % (pkgname, ver, oper))
+        self._dbg(
+            3, "_check_single_pkg_conflict() pkg='%s' ver='%s' oper='%s'" % (
+                pkgname, ver, oper))
         pkg = self._cache[pkgname]
         if pkg.is_installed:
+            assert pkg.installed is not None
             pkgver = pkg.installed.version
         elif pkg.marked_install:
+            assert pkg.candidate is not None
             pkgver = pkg.candidate.version
         else:
             return False
@@ -228,7 +276,7 @@ class DebPackage(object):
         #print "pkgver: %s " % pkgver
         #print "oper: %s " % oper
         if (apt_pkg.check_dep(pkgver, oper, ver) and not
-            self.replaces_real_pkg(pkgname, oper, ver)):
+                self.replaces_real_pkg(pkgname, oper, ver)):
             self._failure_string += _("Conflicts with the installed package "
                                       "'%s'") % pkg.name
             self._dbg(3, "conflicts with installed pkg '%s'" % pkg.name)
@@ -236,6 +284,7 @@ class DebPackage(object):
         return False
 
     def _check_conflicts_or_group(self, or_group):
+        # type: (List[Tuple[str, str, str]]) -> bool
         """Check the or-group for conflicts with installed pkgs."""
         self._dbg(2, "_check_conflicts_or_group(): %s " % (or_group))
         for dep in or_group:
@@ -249,7 +298,7 @@ class DebPackage(object):
                 depname, in_conflict_checking=True)
 
             # check conflicts with virtual pkgs
-            if not depname in self._cache:
+            if depname not in self._cache:
                 # FIXME: we have to check for virtual replaces here as
                 #        well (to pass tests/gdebi-test8.deb)
                 if self._cache.is_virtual_package(depname):
@@ -259,8 +308,8 @@ class DebPackage(object):
                         if self.pkgname == pkg.name:
                             self._dbg(3, "conflict on self, ignoring")
                             continue
-                        if self._check_single_pkg_conflict(pkg.name, ver, 
-                                                           oper):
+                        if self._check_single_pkg_conflict(
+                                pkg.name, ver, oper):
                             self._installed_conflicts.add(pkg.name)
                 continue
             if self._check_single_pkg_conflict(depname, ver, oper):
@@ -269,44 +318,50 @@ class DebPackage(object):
 
     @property
     def conflicts(self):
-        """List of package names conflicting with this package."""
+        # type: () -> List[List[Tuple[str, str, str]]]
+        """List of packages conflicting with this package."""
         key = "Conflicts"
         try:
-            return apt_pkg.parse_depends(self._sections[key])
+            return apt_pkg.parse_depends(self._sections[key], False)
         except KeyError:
             return []
 
     @property
     def depends(self):
-        """List of package names on which this package depends on."""
+        # type: () -> List[List[Tuple[str, str, str]]]
+        """List of packages on which this package depends on."""
         depends = []
         # find depends
         for key in "Depends", "Pre-Depends":
             try:
-                depends.extend(apt_pkg.parse_depends(self._sections[key]))
+                depends.extend(
+                    apt_pkg.parse_depends(self._sections[key], False))
             except KeyError:
                 pass
         return depends
 
     @property
     def provides(self):
+        # type: () -> List[List[Tuple[str, str, str]]]
         """List of virtual packages which are provided by this package."""
         key = "Provides"
         try:
-            return apt_pkg.parse_depends(self._sections[key])
+            return apt_pkg.parse_depends(self._sections[key], False)
         except KeyError:
             return []
 
     @property
     def replaces(self):
+        # type: () -> List[List[Tuple[str, str, str]]]
         """List of packages which are replaced by this package."""
         key = "Replaces"
         try:
-            return apt_pkg.parse_depends(self._sections[key])
+            return apt_pkg.parse_depends(self._sections[key], False)
         except KeyError:
             return []
 
     def replaces_real_pkg(self, pkgname, oper, ver):
+        # type: (str, str, str) -> bool
         """Return True if a given non-virtual package is replaced.
 
         Return True if the deb packages replaces a real (not virtual)
@@ -314,21 +369,26 @@ class DebPackage(object):
         """
         self._dbg(3, "replaces_real_pkg() %s %s %s" % (pkgname, oper, ver))
         pkg = self._cache[pkgname]
+        pkgver = None  # type: Optional[str]
         if pkg.is_installed:
+            assert pkg.installed is not None
             pkgver = pkg.installed.version
         elif pkg.marked_install:
+            assert pkg.candidate is not None
             pkgver = pkg.candidate.version
         else:
             pkgver = None
         for or_group in self.replaces:
             for (name, ver, oper) in or_group:
-                if (name == pkgname and apt_pkg.check_dep(pkgver, oper, ver)):
+                if (name == pkgname and (pkgver is None or
+                        apt_pkg.check_dep(pkgver, oper, ver))):
                     self._dbg(3, "we have a replaces in our package for the "
                                  "conflict against '%s'" % (pkgname))
                     return True
         return False
 
     def check_conflicts(self):
+        # type: () -> bool
         """Check if there are conflicts with existing or selected packages.
 
         Check if the package conflicts with a existing or to be installed
@@ -343,38 +403,47 @@ class DebPackage(object):
         return res
 
     def check_breaks_existing_packages(self):
-        """ 
-        check if installing the package would break exsisting 
+        # type: () -> bool
+        """
+        check if installing the package would break exsisting
         package on the system, e.g. system has:
         smc depends on smc-data (= 1.4)
         and user tries to installs smc-data 1.6
         """
         # show progress information as this step may take some time
         size = float(len(self._cache))
-        steps = max(int(size/50), 1)
+        steps = max(int(size / 50), 1)
         debver = self._sections["Version"]
         debarch = self._sections["Architecture"]
         # store what we provide so that we can later check against that
-        provides = [ x[0][0] for x in self.provides]
+        provides = [x[0][0] for x in self.provides]
         for (i, pkg) in enumerate(self._cache):
-            if i%steps == 0:
-                self._cache.op_progress.update(float(i)/size*100.0)
+            if i % steps == 0:
+                self._cache.op_progress.update(float(i) / size * 100.0)
             if not pkg.is_installed:
                 continue
+            assert pkg.installed is not None
             # check if the exising dependencies are still satisfied
             # with the package
             ver = pkg._pkg.current_ver
             for dep_or in pkg.installed.dependencies:
                 for dep in dep_or.or_dependencies:
                     if dep.name == self.pkgname:
-                        if not apt_pkg.check_dep(debver, dep.relation, dep.version):
+                        if not apt_pkg.check_dep(
+                                debver, dep.relation, dep.version):
                             self._dbg(2, "would break (depends) %s" % pkg.name)
-                            # TRANSLATORS: the first '%s' is the package that breaks, the second the dependency that makes it break, the third the relation (e.g. >=) and the latest the version for the releation
-                            self._failure_string += _("Breaks existing package '%(pkgname)s' dependency %(depname)s (%(deprelation)s %(depversion)s)") % {
-                                'pkgname' : pkg.name, 
-                                'depname' : dep.name, 
-                                'deprelation' : dep.relation, 
-                                'depversion' : dep.version}
+                            # TRANSLATORS: the first '%s' is the package that
+                            # breaks, the second the dependency that makes it
+                            # break, the third the relation (e.g. >=) and the
+                            # latest the version for the releation
+                            self._failure_string += _(
+                                "Breaks existing package '%(pkgname)s' "
+                                "dependency %(depname)s "
+                                "(%(deprelation)s %(depversion)s)") % {
+                                    'pkgname': pkg.name,
+                                    'depname': dep.name,
+                                    'deprelation': dep.relation,
+                                    'depversion': dep.version}
                             self._cache.op_progress.done()
                             return False
             # now check if there are conflicts against this package on
@@ -382,31 +451,48 @@ class DebPackage(object):
             if "Conflicts" in ver.depends_list:
                 for conflicts_ver_list in ver.depends_list["Conflicts"]:
                     for c_or in conflicts_ver_list:
-                        if c_or.target_pkg.name == self.pkgname and c_or.target_pkg.architecture == debarch:
-                            if apt_pkg.check_dep(debver, c_or.comp_type, c_or.target_ver):
-                                self._dbg(2, "would break (conflicts) %s" % pkg.name)
-				# TRANSLATORS: the first '%s' is the package that conflicts, the second the packagename that it conflicts with (so the name of the deb the user tries to install), the third is the relation (e.g. >=) and the last is the version for the relation
-                                self._failure_string += _("Breaks existing package '%(pkgname)s' conflict: %(targetpkg)s (%(comptype)s %(targetver)s)") % {
-                                    'pkgname' : pkg.name, 
-                                    'targetpkg' : c_or.target_pkg.name, 
-                                    'comptype' : c_or.comp_type, 
-                                    'targetver' : c_or.target_ver }
+                        if (c_or.target_pkg.name == self.pkgname and
+                                c_or.target_pkg.architecture == debarch):
+                            if apt_pkg.check_dep(
+                                    debver, c_or.comp_type, c_or.target_ver):
+                                self._dbg(
+                                    2, "would break (conflicts) %s" % pkg.name)
+                                # TRANSLATORS: the first '%s' is the package
+                                # that conflicts, the second the packagename
+                                # that it conflicts with (so the name of the
+                                # deb the user tries to install), the third is
+                                # the relation (e.g. >=) and the last is the
+                                # version for the relation
+                                self._failure_string += _(
+                                    "Breaks existing package '%(pkgname)s' "
+                                    "conflict: %(targetpkg)s "
+                                    "(%(comptype)s %(targetver)s)") % {
+                                        'pkgname': pkg.name,
+                                        'targetpkg': c_or.target_pkg.name,
+                                        'comptype': c_or.comp_type,
+                                        'targetver': c_or.target_ver}
                                 self._cache.op_progress.done()
                                 return False
                         if (c_or.target_pkg.name in provides and
-                            self.pkgname != pkg.name):
-                            self._dbg(2, "would break (conflicts) %s" % provides)
-                            self._failure_string += _("Breaks existing package '%(pkgname)s' that conflict: '%(targetpkg)s'. But the '%(debfile)s' provides it via: '%(provides)s'") % {
-                                'provides' : ",".join(provides),
-                                'debfile'  : self.filename,
-                                'targetpkg' : c_or.target_pkg.name,
-                                'pkgname' : pkg.name }
+                                self.pkgname != pkg.name):
+                            self._dbg(
+                                2, "would break (conflicts) %s" % provides)
+                            self._failure_string += _(
+                                "Breaks existing package '%(pkgname)s' "
+                                "that conflict: '%(targetpkg)s'. But the "
+                                "'%(debfile)s' provides it via: "
+                                "'%(provides)s'") % {
+                                    'provides': ",".join(provides),
+                                    'debfile': self.filename,
+                                    'targetpkg': c_or.target_pkg.name,
+                                    'pkgname': pkg.name}
                             self._cache.op_progress.done()
                             return False
         self._cache.op_progress.done()
         return True
 
     def compare_to_version_in_cache(self, use_installed=True):
+        # type: (bool) -> int
         """Compare the package to the version available in the cache.
 
         Checks if the package is already installed or availabe in the cache
@@ -415,13 +501,19 @@ class DebPackage(object):
         """
         self._dbg(3, "compare_to_version_in_cache")
         pkgname = self._sections["Package"]
+        architecture = self._sections["Architecture"]
+
+        # Arch qualify the package name
+        pkgname = ":".join([pkgname, architecture])
+
         debver = self._sections["Version"]
         self._dbg(1, "debver: %s" % debver)
         if pkgname in self._cache:
-            if use_installed and self._cache[pkgname].installed:
-                cachever = self._cache[pkgname].installed.version
-            elif not use_installed and self._cache[pkgname].candidate:
-                cachever = self._cache[pkgname].candidate.version
+            pkg = self._cache[pkgname]
+            if use_installed and pkg.installed is not None:
+                cachever = pkg.installed.version
+            elif not use_installed and pkg.candidate is not None:
+                cachever = pkg.candidate.version
             else:
                 return self.VERSION_NONE
             if cachever is not None:
@@ -435,33 +527,38 @@ class DebPackage(object):
                     return self.VERSION_OUTDATED
         return self.VERSION_NONE
 
-    def check(self):
+    def check(self, allow_downgrade=False):
+        # type: (bool) -> bool
         """Check if the package is installable."""
         self._dbg(3, "check")
 
         self._check_was_run = True
 
         # check arch
-        if not "Architecture" in self._sections:
+        if "Architecture" not in self._sections:
             self._dbg(1, "ERROR: no architecture field")
             self._failure_string = _("No Architecture field in the package")
             return False
         arch = self._sections["Architecture"]
-        if  arch != "all" and arch != apt_pkg.config.find("APT::Architecture"):
+        if arch != "all" and arch != apt_pkg.config.find("APT::Architecture"):
             if arch in apt_pkg.get_architectures():
                 self._multiarch = arch
                 self.pkgname = "%s:%s" % (self.pkgname, self._multiarch)
                 self._dbg(1, "Found multiarch arch: '%s'" % arch)
             else:
                 self._dbg(1, "ERROR: Wrong architecture dude!")
-                self._failure_string = _("Wrong architecture '%s'") % arch
+                self._failure_string = _("Wrong architecture '%s' "
+                                         "-- Run dpkg --add-architecture to "
+                                         "add it and update afterwards") % arch
                 return False
 
         # check version
-        if self.compare_to_version_in_cache() == self.VERSION_OUTDATED:
+        if (not allow_downgrade and
+            self.compare_to_version_in_cache() == self.VERSION_OUTDATED):
             if self._cache[self.pkgname].installed:
                 # the deb is older than the installed
-                self._failure_string = _("A later version is already installed")
+                self._failure_string = _(
+                    "A later version is already installed")
                 return False
 
         # FIXME: this sort of error handling sux
@@ -471,7 +568,7 @@ class DebPackage(object):
         if not self.check_conflicts():
             return False
 
-        # check if installing it would break anything on the 
+        # check if installing it would break anything on the
         # current system
         if not self.check_breaks_existing_packages():
             return False
@@ -494,20 +591,21 @@ class DebPackage(object):
         return True
 
     def satisfy_depends_str(self, dependsstr):
+        # type: (str) -> bool
         """Satisfy the dependencies in the given string."""
-        return self._satisfy_depends(apt_pkg.parse_depends(dependsstr))
+        return self._satisfy_depends(apt_pkg.parse_depends(dependsstr, False))
 
     def _satisfy_depends(self, depends):
+        # type: (List[List[Tuple[str, str, str]]]) -> bool
         """Satisfy the dependencies."""
         # turn off MarkAndSweep via a action group (if available)
         try:
             _actiongroup = apt_pkg.ActionGroup(self._cache._depcache)
+            _actiongroup  # pyflakes
         except AttributeError:
             pass
         # check depends
         for or_group in depends:
-            #print "or_group: %s" % or_group
-            #print "or_group satified: %s" % self._is_or_group_satisfied(or_group)
             if not self._is_or_group_satisfied(or_group):
                 if not self._satisfy_or_group(or_group):
                     return False
@@ -523,14 +621,17 @@ class DebPackage(object):
 
     @property
     def missing_deps(self):
+        # type: () -> List[str]
         """Return missing dependencies."""
         self._dbg(1, "Installing: %s" % self._need_pkgs)
         if not self._check_was_run:
-            raise AttributeError("property only available after check() was run")
+            raise AttributeError(
+                "property only available after check() was run")
         return self._need_pkgs
 
     @property
     def required_changes(self):
+        # type: () -> Tuple[List[str], List[str], List[str]]
         """Get the changes required to satisfy the dependencies.
 
         Returns: a tuple with (install, remove, unauthenticated)
@@ -539,9 +640,11 @@ class DebPackage(object):
         remove = []
         unauthenticated = []
         if not self._check_was_run:
-            raise AttributeError("property only available after check() was run")
+            raise AttributeError(
+                "property only available after check() was run")
         for pkg in self._cache:
             if pkg.marked_install or pkg.marked_upgrade:
+                assert pkg.candidate is not None
                 install.append(pkg.name)
                 # check authentication, one authenticated origin is enough
                 # libapt will skip non-authenticated origins then
@@ -556,18 +659,20 @@ class DebPackage(object):
 
     @staticmethod
     def to_hex(in_data):
+        # type: (str) -> str
         hex = ""
         for (i, c) in enumerate(in_data):
-            if i%80 == 0:
+            if i % 80 == 0:
                 hex += "\n"
             hex += "%2.2x " % ord(c)
         return hex
 
     @staticmethod
     def to_strish(in_data):
+        # type: (Union[str, Iterable[int]]) -> str
         s = ""
         # py2 compat, in_data is type string
-        if type(in_data) == str:
+        if isinstance(in_data, str):
             for c in in_data:
                 if ord(c) < 10 or ord(c) > 127:
                     s += " "
@@ -581,27 +686,28 @@ class DebPackage(object):
                 else:
                     s += chr(b)
         return s
-        
+
     def _get_content(self, part, name, auto_decompress=True, auto_hex=True):
+        # type: (apt_inst.TarFile, str, bool, bool) -> str
         if name.startswith("./"):
             name = name[2:]
         data = part.extractdata(name)
         # check for zip content
         if name.endswith(".gz") and auto_decompress:
-            io = StringIO(data)
+            io = BytesIO(data)
             gz = gzip.GzipFile(fileobj=io)
-            data = _("Automatically decompressed:\n\n")
+            data = _("Automatically decompressed:\n\n").encode("utf-8")
             data += gz.read()
         # auto-convert to hex
         try:
-            data = unicode(data, "utf-8")
+            return data.decode("utf-8")
         except Exception:
             new_data = _("Automatically converted to printable ascii:\n")
             new_data += self.to_strish(data)
             return new_data
-        return data
 
     def control_content(self, name):
+        # type: (str) -> str
         """ return the content of a specific control.tar.gz file """
         try:
             return self._get_content(self._debfile.control, name)
@@ -609,6 +715,7 @@ class DebPackage(object):
             return ""
 
     def data_content(self, name):
+        # type: (str) -> str
         """ return the content of a specific control.tar.gz file """
         try:
             return self._get_content(self._debfile.data, name)
@@ -616,72 +723,92 @@ class DebPackage(object):
             return ""
 
     def _dbg(self, level, msg):
+        # type: (int, str) -> None
         """Write debugging output to sys.stderr."""
         if level <= self.debug:
-            print >> sys.stderr, msg
+            print(msg, file=sys.stderr)
 
     def install(self, install_progress=None):
+        # type: (Optional[apt.progress.base.InstallProgress]) -> int
         """Install the package."""
+        if self.filename is None:
+            raise apt_pkg.Error("No filename specified")
         if install_progress is None:
             return os.spawnlp(os.P_WAIT, "dpkg", "dpkg", "-i", self.filename)
         else:
             try:
                 install_progress.start_update()
             except AttributeError:
-                install_progress.startUpdate()
+                install_progress.startUpdate()  # type: ignore
             res = install_progress.run(self.filename)
             try:
                 install_progress.finish_update()
             except AttributeError:
-                install_progress.finishUpdate()
+                install_progress.finishUpdate()  # type: ignore
             return res
 
+
 class DscSrcPackage(DebPackage):
     """A locally available source package."""
 
     def __init__(self, filename=None, cache=None):
+        # type: (Optional[str], Optional[apt.Cache]) -> None
         DebPackage.__init__(self, None, cache)
-        self.filename = filename
-        self._depends = []
-        self._conflicts = []
-        self._installed_conflicts = set()
+        self.filename = filename  # type: Optional[str]
+        self._depends = []  # type: List[List[Tuple[str, str, str]]]
+        self._conflicts = []  # type: List[List[Tuple[str, str, str]]]
+        self._installed_conflicts = set()  # type: Set[str]
         self.pkgname = ""
-        self.binaries = []
+        self.binaries = []  # type: List[str]
+        self._sections = {}  # type: Dict[str, str]
         if self.filename is not None:
             self.open(self.filename)
 
     @property
     def depends(self):
+        # type: () -> List[List[Tuple[str, str, str]]]
         """Return the dependencies of the package"""
         return self._depends
 
     @property
     def conflicts(self):
+        # type: () -> List[List[Tuple[str, str, str]]]
         """Return the dependencies of the package"""
         return self._conflicts
 
+    @property
+    def filelist(self):
+        # type: () -> List[str]
+        """Return the list of files associated with this dsc file"""
+        # Files stanza looks like (hash, size, filename, ...)
+        return self._sections['Files'].split()[2::3]
+
     def open(self, file):
+        # type: (str) -> None
         """Open the package."""
         depends_tags = ["Build-Depends", "Build-Depends-Indep"]
         conflicts_tags = ["Build-Conflicts", "Build-Conflicts-Indep"]
-        fobj = open(file)
+        fd = apt_pkg.open_maybe_clear_signed_file(file)
+        fobj = os.fdopen(fd)
         tagfile = apt_pkg.TagFile(fobj)
         try:
             for sec in tagfile:
                 for tag in depends_tags:
-                    if not tag in sec:
+                    if tag not in sec:
                         continue
                     self._depends.extend(apt_pkg.parse_src_depends(sec[tag]))
                 for tag in conflicts_tags:
-                    if not tag in sec:
+                    if tag not in sec:
                         continue
                     self._conflicts.extend(apt_pkg.parse_src_depends(sec[tag]))
                 if 'Source' in sec:
                     self.pkgname = sec['Source']
                 if 'Binary' in sec:
-                    self.binaries = sec['Binary'].split(', ')
-                if 'Version' in sec:
-                    self._sections['Version'] = sec['Version']
+                    self.binaries = [b.strip() for b in
+                                     sec['Binary'].split(',')]
+                for tag in sec.keys():
+                    if tag in sec:
+                        self._sections[tag] = sec[tag]
         finally:
             del tagfile
             fobj.close()
@@ -691,9 +818,13 @@ class DscSrcPackage(DebPackage):
               " ".join(self.binaries))
         self._sections["Description"] = s
         self._check_was_run = False
-        
-    def check(self):
-        """Check if the package is installable.."""
+
+    def check(self, allow_downgrade=False):
+        # type: (bool) -> bool
+        """Check if the package is installable.
+
+        The second parameter is ignored and only exists for compatibility
+        with parent type."""
         if not self.check_conflicts():
             for pkgname in self._installed_conflicts:
                 if self._cache[pkgname]._pkg.essential:
@@ -705,7 +836,9 @@ class DscSrcPackage(DebPackage):
         #        after _satisfy_depends() should probably be done
         return self._satisfy_depends(self.depends)
 
+
 def _test():
+    # type: () -> None
     """Test function"""
     from apt.cache import Cache
     from apt.progress.base import InstallProgress
@@ -713,25 +846,25 @@ def _test():
     cache = Cache()
 
     vp = "www-browser"
-    print "%s virtual: %s" % (vp, cache.is_virtual_package(vp))
+    print("%s virtual: %s" % (vp, cache.is_virtual_package(vp)))
     providers = cache.get_providing_packages(vp)
-    print "Providers for %s :" % vp
+    print("Providers for %s :" % vp)
     for pkg in providers:
-        print " %s" % pkg.name
+        print(" %s" % pkg.name)
 
     d = DebPackage(sys.argv[1], cache)
-    print "Deb: %s" % d.pkgname
+    print("Deb: %s" % d.pkgname)
     if not d.check():
-        print "can't be satified"
-        print d._failure_string
-    print "missing deps: %s" % d.missing_deps
-    print d.required_changes
+        print("can't be satified")
+        print(d._failure_string)
+    print("missing deps: %s" % d.missing_deps)
+    print(d.required_changes)
 
-    print d.filelist
+    print(d.filelist)
 
-    print "Installing ..."
+    print("Installing ...")
     ret = d.install(InstallProgress())
-    print ret
+    print(ret)
 
     #s = DscSrcPackage(cache, "../tests/3ddesktop_0.2.9-6.dsc")
     #s.check_dep()
@@ -739,8 +872,9 @@ def _test():
     #print "Print required changes: ", s.requiredChanges
 
     s = DscSrcPackage(cache=cache)
-    d = "libc6 (>= 2.3.2), libaio (>= 0.3.96) | libaio1 (>= 0.3.96)"
-    print s._satisfy_depends(apt_pkg.parse_depends(d))
+    ds = "libc6 (>= 2.3.2), libaio (>= 0.3.96) | libaio1 (>= 0.3.96)"
+    print(s._satisfy_depends(apt_pkg.parse_depends(ds, False)))
+
 
 if __name__ == "__main__":
     _test()
